use cpal::Sample;
use tracing::trace;

use super::sample::PcmSample;

const TAU: f64 = 2.0 * std::f64::consts::PI;

pub struct TwoStateCarrier {
    pub sample_rate: f64,
    pub freq_low: f64,
    pub freq_high: f64,
    phase_low: f64,
    phase_high: f64,
}

pub trait Carrier {
    fn byte_into_fsk_samples(&mut self, byte: &u8) -> Vec<PcmSample>;
    fn idle(&mut self) -> [PcmSample; 8];
}

impl Carrier for TwoStateCarrier {
    /// turn a u8 into bits &do FSK on the carrier for each, returning samples
    fn byte_into_fsk_samples(&mut self, byte: &u8) -> Vec<PcmSample> {
        //let mut bits: [u8; 8] = [0, 0, 0, 0, 0, 0, 0, 0];
        let mut out: Vec<PcmSample> = vec![];

        // LSb first
        //        bits[0] = (byte     ) & 1;
        //        bits[1] = (byte >> 1) & 1;
        //        bits[2] = (byte >> 2) & 1;
        //        bits[3] = (byte >> 3) & 1;
        //        bits[4] = (byte >> 4) & 1;
        //        bits[5] = (byte >> 5) & 1;
        //        bits[6] = (byte >> 6) & 1;
        //        bits[7] = (byte >> 7) & 1;

        // for each bit in the byte (LSb..=MSb)...
        for bit_place in 1..=8 {
            // if the bit is high...
            if (byte >> bit_place) & 1 == 0 {
                // while the signal hasn't yet completed a single period...
                while (self.phase_low) < self.sample_rate / self.freq_low {
                    // push the next sample of the relative sine wave
                    out.push(PcmSample::F64(
                        (self.phase_low * TAU * self.freq_low
                            / self.sample_rate)
                            .sin(),
                    ));
                    // next sample
                    self.phase_low += 1.0;
                }
                // if phase counter exceeds 1 period, drop it down again
                self.phase_low =
                    self.phase_low % (self.sample_rate / self.freq_low);
            } else if (byte >> bit_place) & 1 == 1 {
                while (self.phase_high) < self.sample_rate / self.freq_high {
                    out.push(PcmSample::F64(
                        (self.phase_high * TAU * self.freq_high
                            / self.sample_rate)
                            .sin(),
                    ));
                    self.phase_high += 1.0;
                }
                self.phase_high =
                    self.phase_high % (self.sample_rate / self.freq_high);
            } else {
                // this keeps happening
                unreachable!("FSK bit math returned a value that isn't 0 or 1");
            }
        }
        //trace!("FSK calculations turned up {:?}", &out);
        out
    }

    fn idle(&mut self) -> [PcmSample; 8] {
        const INIT_VALUE: PcmSample = PcmSample::F64(0.0);
        let mut out: [PcmSample; 8] = [INIT_VALUE; 8];
        for i in 0..=7 {
            out[i] = PcmSample::F64(
                (self.phase_low * TAU * self.freq_low/ self.sample_rate).sin()
            );
            self.phase_low += 1.0;
        }
        self.phase_low = self.phase_low % (self.sample_rate / self.freq_low);
        out
    }
}

impl TwoStateCarrier {
    pub fn new(freq_low: u32, freq_high: u32, sample_rate: u32) -> Self {
        TwoStateCarrier {
            freq_low: freq_low.into(),
            freq_high: freq_high.into(),
            sample_rate: sample_rate.into(),
            phase_low: 0_f64,
            phase_high: 0_f64,
        }
    }
    // setters
    fn freq_low(mut self, freq_low: u32) {
        self.freq_low = freq_low.into();
    }
    fn freq_high(mut self, freq_high: u32) {
        self.freq_high = freq_high.into();
    }
    fn phase_low(mut self, phase_low: u32) {
        self.phase_low = phase_low.into();
    }
    fn phase_high(mut self, phase_high: u32) {
        self.phase_high = phase_high.into();
    }
}
